/*  control.c -- user interaction and control panel window  */

#include <GL/gl.h>
#include <GL/glut.h>
#include "topo.h"

static void reshapeControlPanel(int w, int h);
static void emitControlPanel(void);
static void controlPanelMouseFunc(int button0, int state0, int x0, int y0);
static void move(float step, float turn);
static void setEye(int up);
static void emitString(int x, int y, char *s);



/* control panel size */
#define CONTROL_WIDTH	160
#define CONTROL_HEIGHT	220

/* control panel layout */
#define XLAB		10	/* x position for label */
#define XVAL		100	/* x position for value */
#define LSP		18	/* line spacing (leading) */
#define DSC		4	/* descent below baseline */
#define FONT		GLUT_BITMAP_HELVETICA_12

/* redisplay flags */
#define RC		 1	/* redisplay control window */
#define RO		 2	/* redisplay overhead view */
#define RP		 4	/* redisplay perspective view */



/*  control panel fields  */

struct {
    char cl;		/* keyboard character on left mouse click */
    char cr;		/* keyboard character on right mouse click */
    int variable;	/* is this field variable? */
    int refresh;	/* which windows to refresh */
    int x, y;		/* location in window */
    char label[20];	/* current label value */
} controls[] = {
/*    keyL keyR act  refresh  xposn  yposn    string  */
    { 'c', 'C',  0, RC+RO+RP, XLAB, 1 * LSP, "[C]olor by" },
    { 'c', 'C',  1, RC+RO+RP, XVAL, 1 * LSP, "" },
    { 'r', 'R',  0, RC+RO+RP, XLAB, 2 * LSP, "[R]efine by" },
    { 'r', 'R',  1, RC+RO+RP, XVAL, 2 * LSP, "" },
    { '+', '-',  0, RC+RO+RP, XLAB, 3 * LSP, "Threshold[+/-]" },
    { '+', '-',  1, RC+RO+RP, XVAL, 3 * LSP, "" },
    { 'e', 'E',  0, RC+RO+RP, XLAB, 5 * LSP, "[E]ye" },
    { 'e', 'E',  1, RC+RO+RP, XVAL, 5 * LSP, "" },
    { '2', '@',  0, RC+RO   , XLAB, 6 * LSP, "[2]d Mesh" },
    { '2', '@',  1, RC+RO   , XVAL, 6 * LSP, "" },
    { '3', '#',  0, RC   +RP, XLAB, 7 * LSP, "[3]d Mesh" },
    { '3', '#',  1, RC   +RP, XVAL, 7 * LSP, "" },
    { 'd', 'D',  0,       RP, XLAB, 9 * LSP, "[D]etails" },
    { 'p', 'P',  0,        0, XLAB,10 * LSP, "[P]rint" },
    { 'q', 'Q',  0,        0, XLAB,11 * LSP, "[Q]uit" }
};

#define NUMBER_OF_CONTROLS (sizeof(controls) / sizeof(controls[0]))



/*  openCtlWin() -- open control panel window  */

void openCtlWin(void)
{
    char wname[200];

    sprintf(wname, "%s:C", dmname);
    glutInitDisplayMode(GLUT_RGB | GLUT_SINGLE);
    glutInitWindowSize(CONTROL_WIDTH, CONTROL_HEIGHT);
    ctlWin = glutCreateWindow(wname);
    glutSetWindowTitle(wname);
    glClearColor(1.0, 1.0, 1.0, 1.0);
    glColor3f(0, 0, 0);

    glutMouseFunc(controlPanelMouseFunc);
    glutDisplayFunc(emitControlPanel);
    glutReshapeFunc(reshapeControlPanel);
    glutKeyboardFunc(keyfunc);
    glutSpecialFunc(specfunc);
}



/*  reshapeControlPanel(w, h) -- set rendering parameters for control panel  */

static void reshapeControlPanel(int w, int h)
{
    glViewport(0, 0, w, h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0, w, 0, h);
    glScalef(1, -1, 1);
    glTranslatef(0, -h, 0);
    glMatrixMode(GL_MODELVIEW);
}



/*  emitControlPanel() -- display contents of control panel  */

static void emitControlPanel(void)
{
    int i;
    char *s;

    glClear(GL_COLOR_BUFFER_BIT);

    for(i = 0; i < NUMBER_OF_CONTROLS; ++i) {
	if (controls[i].x == XVAL) {
	    s = controls[i].label;

	    switch(controls[i].cl) {

		case 'r':
		    switch(surfMethod) {
			case ERR:   strcpy(s, "Error");		break;
			case DIST:  strcpy(s, "Distance");	break;
			case HEAP:  strcpy(s, "Heap");		break;
		    }
		    break;

		case '+':
		    switch(surfMethod) {
			case ERR:   sprintf(s, "%8.6f", errThresh);   break;
			case DIST:  sprintf(s, "%4.1f", stepThresh);  break;
			case HEAP:  sprintf(s, "%d", heapThresh);     break;
		    }
		    break;

		case 'c':
		    switch(reliefMethod) {
			case T_CONTOURS:  strcpy(s, "Contours");  break;
			case T_GRAY:      strcpy(s, "Relief");    break;
			case T_RGB:       strcpy(s, "PPM file");  break;
		    }
		    break;

		case 'e':  strcpy(s, eyeUp ? "Up" : "Down");	break;

		case '2':
		    switch(twoDMesh) {
			case NONE2D:  strcpy(s, "Off");		break;
			case WIRE2D:  strcpy(s, "Outlined");	break;
			case FILL2D:  strcpy(s, "Filled");	break;
		    }
		    break;

		case '3':  strcpy(s, threeDMesh ? "On" : "Off");  break;

	    }
	}
    }

    for(i = 0; i < NUMBER_OF_CONTROLS; ++i) {
	if (controls[i].variable) {
	    glColor3f(1.0, 0.0, 0.0);
	} else {
	    glColor3f(0.0, 0.0, 0.0);
	}

	emitString(controls[i].x, controls[i].y, controls[i].label);

#ifdef SHOWCONTROLRECTANGLE
	glBegin(GL_LINE_LOOP);
	glColor3f(0.0, 0.0, 0.0);
	glVertex2i(2, controls[i].y + DSC);
	glVertex2i(CONTROL_WIDTH - 3, controls[i].y + DSC);
	glVertex2i(CONTROL_WIDTH - 3, controls[i].y - LSP + DSC);
	glVertex2i(2, controls[i].y - LSP + DSC);
	glEnd();
#endif

    }
    glFlush();
}



/*  emitString(x, y, s) -- display a string  */

static void emitString(int x, int y, char *s)
{
    int len, i;

    glRasterPos2i(x, y);
    len = strlen(s);
    for(i = 0; i < len; ++i) {
	glutBitmapCharacter(FONT, s[i]);
    }
}



/*  controlPanelMouseFunc(btn, state, x, y) -- handle click in control window */

static void controlPanelMouseFunc(int button0, int state0, int x0, int y0)
{
    int i;

    if (state0 != GLUT_UP)
	return;

    for(i = 0; i < NUMBER_OF_CONTROLS; ++i) {
	if (y0 < controls[i].y + DSC && y0 > controls[i].y + DSC - LSP) {
	    if (button0 == GLUT_LEFT_BUTTON)
	        keyfunc(controls[i].cl, x0, y0);
	    else if (button0 == GLUT_RIGHT_BUTTON)
		keyfunc(controls[i].cr, x0, y0);
	    break;
	}
    }
}


/*  keyfunc(key, x, y) -- handle keypress in any window  */

void keyfunc(unsigned char key, int x, int y)
/*ARGSUSED*/
{
    int i, sm;
    float et;
    int sign = 1;
    int refresh = -1;
    extern int noRefresh;


    switch(key) {

    case 'L':
      noRefresh = 1;
      /* NO BREAK */
    case 'l':
      doloop();
      noRefresh = 0;
      break;
      
    case 'r':
	switch(surfMethod) {
	    case ERR:   surfMethod = DIST;  break;
	    case DIST:  surfMethod = HEAP;  break;
	    case HEAP:  surfMethod = ERR;   break;
	}
	viewChanged();
	break;
    case 'R':
	switch(surfMethod) {
	    case DIST:  surfMethod = ERR;   break;
	    case HEAP:  surfMethod = DIST;  break;
	    case ERR:   surfMethod = HEAP;  break;
	}
	viewChanged();
	break;

    case '-':
	sign = -sign;
	/*NOBREAK*/
    case '+':
	switch(surfMethod) {
	    case ERR:   errThresh  += sign * errThreshInc;   break;
	    case DIST:  stepThresh += sign * stepThreshInc;  break;
	    case HEAP:  heapThresh += sign * heapThreshInc;  break;
	}
	viewChanged();
	key = 't';
	break;

    case 'c':
	switch(reliefMethod) {
	case T_CONTOURS:
	    reliefMethod = T_GRAY;
	    break;
	case T_GRAY:
	    if (cgrid)
		reliefMethod = T_RGB;
	    else
		reliefMethod = T_CONTOURS;
	    break;
	case T_RGB:
	    reliefMethod = T_CONTOURS;
	    break;
	}
	break;

    case 'C':
	switch(reliefMethod) {
	case T_GRAY:
	    reliefMethod = T_CONTOURS;
	    break;
	case T_CONTOURS:
	    if (cgrid)
		reliefMethod = T_RGB;
	    else
		reliefMethod = T_GRAY;
	    break;
	case T_RGB:
	    reliefMethod = T_GRAY;
	    break;
	}
	break;

    case '2':
	switch(twoDMesh) {
	    case NONE2D:  twoDMesh = WIRE2D;	break;
	    case WIRE2D:  twoDMesh = FILL2D;	break;
	    case FILL2D:  twoDMesh = NONE2D;	break;
	}
	break;

    case '@':
	switch(twoDMesh) {
	    case WIRE2D:  twoDMesh = NONE2D;	break;
	    case FILL2D:  twoDMesh = WIRE2D;	break;
	    case NONE2D:  twoDMesh = FILL2D;	break;
	}
	break;

    case 'D':
    case 'd':
	sm = surfMethod;
	et = errThresh;
	surfMethod = ERR;
	errThresh = 0.0;
	viewChanged();
	buildSurface();
	surfMethod = sm;
	errThresh = et;
    	break;

    case 'E':
    case 'e':  setEye(!eyeUp);			break;

    case '#':
    case '3':  threeDMesh = !threeDMesh;	break;

    case 'P':
    case 'p':  printRTIN(hierarchy);		break;

    case 'Q':
    case 'q':  exit(0);				break;

    case ' ':  move(+XMAX / 5, 0);		break;
    case '\b': move(-XMAX / 5, 0);		break;

    default:  return;	/* unrecognized; no further processing */
    }

    /* search control panel table for refresh flags */
    for(i = 0; i < NUMBER_OF_CONTROLS; ++i) {
	if (controls[i].cl == key || controls[i].cr == key) {
	    refresh = controls[i].refresh;
	    break;
	}
    }

    /* refresh any necessary displays according to the flags set */
    if (refresh & RC) {
	glutSetWindow(ctlWin);
	glutPostRedisplay();
    }
    if (refresh & RO) {
	glutSetWindow(ovhWin);
	glutPostRedisplay();
    }
    if (refresh & RP) {
	glutSetWindow(pspWin);
	glutPostRedisplay();
    }
}



/*  specfunc(key, x, y) -- handle function key or other special key  */

void specfunc(int key, int x, int y)
/*ARGSUSED*/
{
    switch (key) {
	case GLUT_KEY_UP:		move(+XMAX / 20, 0);	break;
	case GLUT_KEY_DOWN:		move(-XMAX / 20, 0);	break;
	case GLUT_KEY_LEFT:		move(0.0, +15.0);	break;
	case GLUT_KEY_RIGHT:		move(0.0, -15.0);	break;
	case GLUT_KEY_PAGE_UP:		setEye(1);		break;
	case GLUT_KEY_PAGE_DOWN:	setEye(0);		break;
	default:			/* nothing */		break;
    }
}



/*  move(step, turn) -- step forward and swivel gaze */

/*  if the shift key is down, scale both argument values by 0.25  */
/*  (call only from within a callback: it calls glutGetModifiers) */

static void move(float step, float turn)
{
    float dx, dy, d, a;

    if (glutGetModifiers() & GLUT_ACTIVE_SHIFT) {
	step *= 0.25;
	turn *= 0.25;
    }

    dx = refWx - eyeWx;
    dy = refWy - eyeWy;
    d = sqrt(dx * dx + dy * dy);	/* current distance */
    a = atan2(dy, dx);			/* current view angle */

    eyeWx += step * cos(a);		/* new eye location */
    eyeWy += step * sin(a);
    refWx += step * cos(a);		/* also translate reference point */
    refWy += step * sin(a);

    a += .017453293 * turn;		/* new view angle (in radians) */
    refWx = eyeWx + d * cos(a);		/* new reference point */
    refWy = eyeWy + d * sin(a);

    viewChanged();			/* note redisplay needed */
}



/*  setEye(up) -- set eye height for current viewpoint  */
/*  calls PostRedisplay if up/down flag has changed  */

static void setEye(int up)
{
    if (up != eyeUp) {
	eyeUp = up;		/* save new value if different */
	viewChanged();		/* note that view changed */
	glutSetWindow(ctlWin);	/* rewrite control window to show new value */
	glutPostRedisplay();
    }
}
