/*  ctg2ppm -- convert CTG file to PPM file with coordinates in comments  */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "basic.h"
#include "ppmops.h"



/* classifications and colors */

typedef struct {		/* info about a CTG classification */
    int code;
    char *label;
    float red, green, blue;
} CTGInfo;

CTGInfo *ctgInfo(int key);

CTGInfo ctgKey[] = {
/* 1 Urban or Built-Up Land */
    { 11, "Residential",			0.25, 0.25, 0.25 },
    { 12, "Commercial Services",		0.35, 0.35, 0.35 },
    { 13, "Industrial",				0.45, 0.45, 0.45 },
    { 14, "Transportation, Communications",	1.00, 0.00, 0.00 },
    { 15, "Industrial and Commercial",		0.55, 0.55, 0.55 },
    { 16, "Mixed Urban or Built-Up Land",	0.5, 0.5, 0.5 },
    { 17, "Other Urban or Built-Up Land",	0.5, 0.5, 0.5 },
/* 2 Agricultural Land */
    { 21, "Cropland and Pasture",		0.50, 0.50, 0.0 },
    { 22, "Orchards, Groves, Vineyards, Nurseries", 0.50, 0.50, 0.1 },
    { 23, "Confined Feeding Operations",	0.50, 0.50, 0.2 },
    { 24, "Other Agricultural Land",		0.5, 0.5, 0.3 },
/* 3 Rangeland */
    { 31, "Herbaceous Rangeland",		0.81, 0.68, 0.40 },
    { 32, "Shrub and Brush Rangeland",		0.87, 0.81, 0.50 },
    { 33, "Mixed Rangeland",			0.84, 0.75, 0.45 },
/* 4 Forest Land */
    { 41, "Deciduous Forest Land",		0.0, 0.45, 0.18 },
    { 42, "Evergreen Forest Land",		0.16, 0.45, 0.35 },
    { 43, "Mixed Forest Land",			0.08, 0.40, 0.20 },
/* 5 Water */
    { 51, "Streams and Canals",			0.0, 0.45, 0.77 },
    { 52, "Lakes",				0.0, 0.45, 0.77 },
    { 53, "Reservoirs",				0.0, 0.45, 0.77 },
    { 54, "Bays and Estuaries",			0.0, 0.45, 0.77 },
/* 6 Wetland */
    { 61, "Forested Wetlands",			0.23, 0.64, 0.49 },
    { 62, "Nonforested Wetlands",		0.23, 0.6, 0.55 },
/* 7 Barren Land */
    { 71, "Dry Salt Flats",			0.9, 0.9, 0.9 },
    { 72, "Beaches",				0.9, 0.9, 0.8 },
    { 73, "Sandy Areas Other than Beaches",	0.9, 0.9, 0.7 },
    { 74, "Bare Exposed Rock",			0.9, 0.8, 0.7 },
    { 75, "Strip Mines, Quarries, and Gravel Pits", 0.8, 0.8, 0.7 },
    { 76, "Transitional Areas",			0.8, 0.7, 0.7 },
    { 77, "Mixed Barren Land",			0.9, 0.9, 0.6 },
/* 8 Tundra */
    { 81, "Shrub and Brush Tundra",		0.7, 0.7, 0.7 },
    { 82, "Herbaceous Tundra",			0.7, 0.7, 0.7 },
    { 83, "Bare Ground",			0.7, 0.7, 0.7 },
    { 84, "Wet Tundra",				0.7, 0.7, 0.7 },
    { 85, "Mixed Tundra",			0.7, 0.7, 0.7 },
/* 91 Perennial Snowfields */
    { 92, "Glaciers",				1.0, 1.0, 1.0 },
/* sentinel & default value -- code 0, no label string, dark red */
    {  0, 0, 0.4, 0.0, 0.0 }
};



/*  CTG data structure  */

typedef struct {
    int number_of_rows;
    int number_of_cells;
    int number_of_columns;
    int cell_size;
    int num_overlays;
    int map_type;
    int zone_number;
    int projection_code;
    int scale;
    int source_date;
    int min_col_index, min_row_index;
    int max_col_index, max_row_index;
    int sw_col_index, sw_row_index;
    int nw_col_index, nw_row_index;
    int nc_col_index, nc_row_index;
    int ne_col_index, ne_row_index;
    int se_col_index, se_row_index;
    int sc_col_index, sc_row_index;
    int sw_lat, sw_lon;
    int nw_lat, nw_lon;
    int nc_lat, nc_lon;
    int ne_lat, ne_lon;
    int se_lat, se_lon;
    int sc_lat, sc_lon;
    int nw_utm_easting, nw_utm_northing;
    int creation_date;
    char title[65];
    char *data;
} Ctg;

#define CTGRef(cp, i, j)  ((cp)->data[(j)*(cp)->number_of_columns + (i)])



void initColors(void);
void readCTG(FILE *fp, Ctg *cp);



/*  main  */

int main(int argc, char *argv[])
{
    int i, j;
    FILE *fp;
    Ctg ctg, *cp;
    CTGInfo *c;
    UV swUV, nwUV, neUV, seUV;
    UV swXY, nwXY, neXY, seXY;

    if (argc > 2) {
	fprintf(stderr, "Usage: %s [filename]\n", argv[0]);
	exit(1);
    }

    if (argc < 2)
	fp = stdin;
    else {
	fp = fopen(argv[1], "r");
	if (fp == NULL) {
	    fprintf(stderr, "Can't open %s for read\n", argv[1]);
	    exit(1);
	}
    }

    initColors();		/* load mapping table */

    cp = &ctg;
    readCTG(fp, cp);

    swXY.u = cp->sw_col_index - cp->min_col_index;
    swXY.v = cp->max_row_index - cp->sw_row_index;
    nwXY.u = cp->nw_col_index - cp->min_col_index;
    nwXY.v = cp->max_row_index - cp->nw_row_index;
    neXY.u = cp->ne_col_index - cp->min_col_index;
    neXY.v = cp->max_row_index - cp->ne_row_index;
    seXY.u = cp->se_col_index - cp->min_col_index;
    seXY.v = cp->max_row_index - cp->se_row_index;

    swUV.u = cp->sw_lon;
    swUV.v = cp->sw_lat;
    nwUV.u = cp->nw_lon;
    nwUV.v = cp->nw_lat;
    neUV.u = cp->ne_lon;
    neUV.v = cp->ne_lat;
    seUV.u = cp->se_lon;
    seUV.v = cp->se_lat;

    ppmWriteHeader(stdout, cp->number_of_columns, cp->number_of_rows,
		    swUV, nwUV, neUV, seUV,
		    swXY, nwXY, neXY, seXY);

    for (i = 0; i < cp->number_of_rows; i++) {
	for (j = 0; j <cp->number_of_columns; j++) {
	    c = ctgInfo(*(cp->data + i * cp->number_of_columns + j));
	    printf("%c%c%c", (unsigned char)(c->red * 255.0),
			    (unsigned char)(c->green * 255.0),
			    (unsigned char)(c->blue * 255.0));
	}
    }
    return 0;
}



/*  initColors()  -- initialize color mapping table  */

#define CTGMAX 100			/* maximum key value plus 1 */
CTGInfo *ctgMap[CTGMAX];		/* map from key to label and color */

void initColors(void)
{
    int i;
    CTGInfo *p;

    /* preload with pointers to sentinel/filler entry */
    i = sizeof(ctgKey) / sizeof(ctgKey[0]);	/* number of entries */
    p = ctgKey + i - 1;				/* pointer to last entry */
    for (i = 0; i < CTGMAX; i++)
	ctgMap[i] = p;

    /* load recognized values from table above */
    for (p = ctgKey; p->code != 0; p++)
	ctgMap[p->code] = p;
}



/*  ctgInfo(key) -- return pointer to Info struct for key  */

CTGInfo *ctgInfo(int key)
{
    if (key >= 0 && key < CTGMAX)	/* if in range */
	return ctgMap[key];		/* return corresponding entry */
    else
	return ctgMap[0];		/* else return first entry */
}



/* CTG processing code */


/*--------------------------getCTGInt()--------------------------*/

int getCTGInt(char *file_record, int start_pos, int end_pos) {
    char tmp_str[100];

    memcpy(tmp_str, file_record + start_pos - 1, end_pos - start_pos + 1);
    tmp_str[end_pos-start_pos + 1] = 0;

    return atoi(tmp_str);
}

/*--------------------------getCTGLatLon()--------------------------*/

int getCTGLatLon(char *file_record, int start_pos, int end_pos) {
    char tmp_str[100];
    int degrees, minutes, seconds;

    memcpy(tmp_str, file_record + start_pos - 1, 10);
    tmp_str[10] = 0;
    seconds = atoi(tmp_str + 8);
    tmp_str[8] = 0;
    minutes = atoi(tmp_str + 6);
    tmp_str[6] = 0;
    degrees = atoi(tmp_str);

    return seconds + (minutes + degrees * 60) * 60;
}

/*--------------------------set_ctg_element()--------------------------*/

void set_ctg_element(Ctg *cp, int row, int column, int element) {

    *(cp->data + row * cp->number_of_columns + column) = element;
}

/*--------------------------readCTG()--------------------------*/

void readCTG(FILE *fp, Ctg *cp) {
    char file_record[81];
    int i, j;
    int records_read, percent_recs_read;
    int x, y, column, row;

    /* RECORD 1 */
    if (!fread(file_record, 80, 1, fp)) {
	fprintf(stderr, "Error reading CTG file... aborting.\n");
	exit(1);
    }
    cp->number_of_rows = getCTGInt(file_record, 1, 10);
    cp->number_of_cells = getCTGInt(file_record, 11, 20) / 2;
    cp->number_of_columns = getCTGInt(file_record, 21, 30);

    cp->cell_size = getCTGInt(file_record, 36, 40);
    cp->num_overlays = getCTGInt(file_record, 41, 45);
    cp->map_type = getCTGInt(file_record, 46, 50);
    cp->zone_number = getCTGInt(file_record, 51, 55);
    cp->projection_code = getCTGInt(file_record, 56, 60);
    cp->scale = getCTGInt(file_record, 61, 70);
    cp->source_date = getCTGInt(file_record, 61, 70);

    /* RECORD 2 */
    if (!fread(file_record, 80, 1, fp)) {
	fprintf(stderr, "Error reading CTG file... aborting.\n");
	exit(1);
    }
    cp->min_col_index = getCTGInt(file_record, 1, 5);
    cp->min_row_index = getCTGInt(file_record, 6, 10);
    cp->max_col_index = getCTGInt(file_record, 11, 15);
    cp->max_row_index = getCTGInt(file_record, 16, 20);
    cp->sw_col_index = getCTGInt(file_record, 21, 25);
    cp->sw_row_index = getCTGInt(file_record, 26, 30);
    cp->nw_col_index = getCTGInt(file_record, 31, 35);
    cp->nw_row_index = getCTGInt(file_record, 36, 40);
    cp->nc_col_index = getCTGInt(file_record, 41, 45);
    cp->nc_row_index = getCTGInt(file_record, 46, 50);
    cp->ne_col_index = getCTGInt(file_record, 51, 55);
    cp->ne_row_index = getCTGInt(file_record, 56, 60);
    cp->se_col_index = getCTGInt(file_record, 61, 65);
    cp->se_row_index = getCTGInt(file_record, 66, 70);
    cp->sc_col_index = getCTGInt(file_record, 71, 75);
    cp->sc_row_index = getCTGInt(file_record, 76, 80);

    /* RECORD 3 */
    if (!fread(file_record, 80, 1, fp)) {
	fprintf(stderr, "Error reading CTG file... aborting.\n");
	exit(1);
    }
    cp->sw_lat = getCTGLatLon(file_record, 1, 10);
    cp->sw_lon = getCTGLatLon(file_record, 11, 20);
    cp->nw_lat = getCTGLatLon(file_record, 21, 30);
    cp->nw_lon = getCTGLatLon(file_record, 31, 40);
    cp->nc_lat = getCTGLatLon(file_record, 41, 50);
    cp->nc_lon = getCTGLatLon(file_record, 51, 60);
    cp->ne_lat = getCTGLatLon(file_record, 61, 70);
    cp->ne_lon = getCTGLatLon(file_record, 71, 80);

    /* RECORD 4 */
    if (!fread(file_record, 80, 1, fp)) {
	fprintf(stderr, "Error reading CTG file... aborting.\n");
	exit(1);
    }
    cp->se_lat = getCTGLatLon(file_record, 1, 10);
    cp->se_lon = getCTGLatLon(file_record, 11, 20);
    cp->sc_lat = getCTGLatLon(file_record, 21, 30);
    cp->sc_lon = getCTGLatLon(file_record, 31, 40);
    /* Fix unsigned longitudes (grrr...) */
    if (cp->zone_number < 31) {
	cp->sw_lon = -ABS(cp->sw_lon);
	cp->nw_lon = -ABS(cp->nw_lon);
	cp->nc_lon = -ABS(cp->nc_lon);
	cp->ne_lon = -ABS(cp->ne_lon);
	cp->se_lon = -ABS(cp->se_lon);
	cp->sc_lon = -ABS(cp->sc_lon);
    }
    cp->nw_utm_easting = getCTGInt(file_record, 41, 50);
    cp->nw_utm_northing = getCTGInt(file_record, 51, 60);
    cp->creation_date = getCTGInt(file_record, 61, 70);

    /* RECORD 5 */
    if (!fread(file_record, 80, 1, fp)) {
	fprintf(stderr, "Error reading CTG file... aborting.\n");
	exit(1);
    }

    memcpy(cp->title, file_record, 64);
    cp->title[64] = 0;

    cp->data = malloc(cp->number_of_rows * cp->number_of_columns);
    if (!cp->data){
	fprintf(stderr, "Cannot allocate memory for CTG\n");
	exit(1);
    }

    for (i = 0; i < cp->number_of_rows; i++) {
	for (j = 0; j < cp->number_of_columns; j++) {
	    set_ctg_element(cp, i, j, 0);
	}
    }

    records_read = 0;
    percent_recs_read = -1;
    while (fread(file_record, 80, 1, fp)) {
	x = getCTGInt(file_record, 4, 11);
	y = getCTGInt(file_record, 12, 19);
	column = (x - cp->nw_utm_easting)/cp->cell_size + 1 - cp->min_col_index;
	row = (cp->nw_utm_northing - y)/cp->cell_size + 1 - cp->min_row_index;
	if (column >= cp->number_of_columns || column < 0) {
	    fprintf(stderr, "Easting out of range (%d meters)\n", x);
	    exit(1);
	}
	if (row >= cp->number_of_rows || row < 0) {
	    fprintf(stderr, "Northing out of range (%d meters)\n", y);
	    exit(1);
	}
	set_ctg_element(cp, row, column, getCTGInt(file_record, 21, 30));
	records_read++;
	if (percent_recs_read != records_read * 100 / cp->number_of_cells) {
	    percent_recs_read = (records_read * 100 / cp->number_of_cells);
	    fprintf(stderr, "\rReading CTG file... %d%%", percent_recs_read);
	    fflush(stdout);
	}
    }

    fprintf(stderr, "\n");
}
