/*  dem.c -- functions for reading USGS Digital Elevation Model files  */

/* Will Evans (adapted from Scott Andrews) */



#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "dem.h"

static int tryRead(FILE *fp, Dem *dp, int trial);
static int getDEMInt(FILE *fp);
static int getDEMInt6(FILE *fp);
static double getDEMDouble12(FILE *fp);
static double getDEMDouble24(FILE *fp);



void readDEM(FILE *fp, Dem *dp)
{
    if (tryRead(fp, dp, 1))
	return;

    if (fseek(fp, 0L, SEEK_SET) < 0) {
	fprintf(stderr, "Cannot rewind input file.\n");
	exit(1);
    }

    if (tryRead(fp, dp, 2))
	return;

    fprintf(stderr, "Not standard DEM format.\n");
    exit(1);
}



static int tryRead(FILE *fp, Dem *dp, int trial)
{
    int i;
    int col;
    int row_id, col_id;
    int colsInProfiles, numProfiles;
    UV min, max;
    int firstIndex;
    float scaleElevation;

    /* A records */

    /* A 1 */
    if (!fgets(dp->name, 145, fp)) {
	fprintf(stderr, "Empty input file.\n");
	exit(1);
    }

    /* A 2 */
    dp->level = getDEMInt6(fp);

    /* A 3 */
    dp->elevation_pattern = getDEMInt6(fp);

    /* A 4 */
    dp->reference_system = getDEMInt6(fp);

    /* A 5 */
    dp->zone = getDEMInt6(fp);

    /* A 6 */
    for (i = 0; i < 15; i++) {
	dp->map_projection_parameter[i] = getDEMDouble24(fp);
    }

    /* A 7 */
    dp->ground_unit = getDEMInt6(fp);

    /* A 8 */
    dp->elevation_unit = getDEMInt6(fp);
    if (dp->elevation_unit == 1) {
	scaleElevation = METERS_PER_FEET;
    } else {
	scaleElevation = 1.0;
    }

    /* A 9 */
    dp->n_polygon = getDEMInt6(fp);

    /* A 10 */
    for (i = 0; i < 4; i++) {
	dp->corner[i].u = getDEMDouble24(fp);
	dp->corner[i].v = getDEMDouble24(fp);
	if (i == 0) {
	    min.u = dp->corner[i].u;
	    max.u = dp->corner[i].u;
	    min.v = dp->corner[i].v;
	    max.v = dp->corner[i].v;
	} else {
	    min.u = MIN(min.u, dp->corner[i].u);
	    max.u = MAX(max.u, dp->corner[i].u);
	    min.v = MIN(min.v, dp->corner[i].v);
	    max.v = MAX(max.v, dp->corner[i].v);
	}
    }

    /* A 11 */
    dp->minElevation = getDEMDouble24(fp) * scaleElevation;
    dp->maxElevation = getDEMDouble24(fp) * scaleElevation;

    /* A 12 */
    dp->axis_angle = getDEMDouble24(fp);
    if (dp->axis_angle != 0.0) {
	fprintf(stderr, "Assuming base angle of 0 (real angle is %lf)\n",
			dp->axis_angle);
    }

    /* A 13 */
    dp->accuracy = getDEMInt6(fp);

    /* A 14 */
    for (i = 0; i < 3; i++) {
	dp->resolution[i] = getDEMDouble12(fp);
    }

    /* A 15 */
    colsInProfiles = getDEMInt6(fp);
    if (colsInProfiles != 1) {
	fprintf(stderr, "Cannot handle more than one column profiles\n");
	exit(1);
    }
    numProfiles = getDEMInt6(fp);
    if (numProfiles > MAXCOLS) {
	fprintf(stderr, "Too many rows in data set\n");
	exit(1);
    }

    /* A remaining (ignored) */
    for (i = 864; i < 1024; i++)
	if (getc(fp) == '\n')
	    break;

    /* B records */

    dp->cols = numProfiles;
    dp->rows = 1 + (max.v - min.v)/dp->resolution[1];

#ifdef DEBUG
    fprintf(stderr, "Rectangle is %d cols by %d rows\n", dp->cols, dp->rows);
#endif

    if (dp->rows > MAXROWS) {
	fprintf(stderr, "Too many rows in DEM\n");
	exit(1);
    }
    for (col = 0; col < numProfiles; ++col) {
	for (i = 0; i < dp->rows; ++i) {
	    dp->data[col][i] = 0;
	}

	/* B 1 */
	row_id = getDEMInt(fp);
	col_id = getDEMInt(fp);
	if (col_id != col + 1 || row_id != 1) {
#ifdef DEBUG
	    fprintf(stderr, "Profile %d claims to be %d\n", col + 1, col_id);
#endif
	    return 0;
	}

	/* B 2 */
	dp->rowsInProfile[col] = getDEMInt6(fp);
	dp->colsInProfile[col] = getDEMInt6(fp);

	/* B 3 */
	dp->firstUVInProfile[col].u = getDEMDouble24(fp);
	dp->firstUVInProfile[col].v = getDEMDouble24(fp);
	firstIndex = (dp->firstUVInProfile[col].v - min.v)/dp->resolution[1];

	/* B 4 */
	dp->firstElevationInProfile[col] = getDEMDouble24(fp) * scaleElevation;

	/* B 5 */
	dp->minElevationInProfile[col] = getDEMDouble24(fp) * scaleElevation;
	dp->maxElevationInProfile[col] = getDEMDouble24(fp) * scaleElevation;

	/* B 6 */
#ifdef DEBUG
	fprintf(stderr, "Profile %d starts with row (%lf - %lf)/%lf = %d\n",
		col + 1, dp->firstUVInProfile[col].v, min.v, dp->resolution[1],
		firstIndex);
#endif

	for (i = firstIndex; i < firstIndex + dp->rowsInProfile[col]; i++) {
	    if (i >= dp->rows) {
		fprintf(stderr, "Too many rows in DEM profile %d\n", col + 1);
		exit(1);
	    }

	    if (trial == 1)
		dp->data[col][i] = getDEMInt(fp) * scaleElevation;
	    else
		dp->data[col][i] = getDEMInt6(fp) * scaleElevation;
	}

    }

    /* C record */

    if (dp->accuracy != 0) {

	/* C 1 */
	dp->C2 = getDEMInt6(fp);

	/* C 2 */
	dp->C2x = getDEMInt6(fp);
	dp->C2y = getDEMInt6(fp);
	dp->C2z = getDEMInt6(fp);

	/* C 3 */
	dp->C2sample_size = getDEMInt6(fp);

	/* C 4 */
	dp->C5 = getDEMInt6(fp);

	/* C 5 */
	dp->C5x = getDEMInt6(fp);
	dp->C5y = getDEMInt6(fp);
	dp->C5z = getDEMInt6(fp);

	/* C 6 */
	dp->C5sample_size = getDEMInt6(fp);
    }

    return 1;
}



static void getDEMNoCR(int n, char *buf, FILE *fp)
{
    int i;

    for (i = 0; i < n; ++i) {
	buf[i] = fgetc(fp);
	if (buf[i] == '\n') {
	    --i;
	}
    }
    buf[n] = '\0';

}



static void getDEMD2E(int n, char *buf, FILE *fp)
{
    int i;

    for (i = 0; i < n; ++i) {
	buf[i] = fgetc(fp);
	if (buf[i] == 'D') {
	    buf[i] = 'E';
	}
    }
    buf[n] = '\0';
}



/* ? characters */
static int getDEMInt(FILE *fp)
{
    int i;

    fscanf(fp, "%d", &i);
    return i;
}



/* 6 characters */
static int getDEMInt6(FILE *fp)
{
    char buf[7];

    getDEMNoCR(6, buf, fp);
    return atoi(buf);
}



/* 12 characters */
static double getDEMDouble12(FILE *fp)
{
    char buf[13];

    getDEMD2E(12, buf, fp);
    return atof(buf);
}



/* 24 characters */
static double getDEMDouble24(FILE *fp)
{
    char buf[25];

    getDEMD2E(24, buf, fp);
    return atof(buf);
}
