/*  dminfo -- print info from DM file header  */

/*
 *  usage:  dminfo [-v] [file...]
 *
 *	-v	print verbose report
 *
 *  Normally prints grid size and latitude/longitude of two corners.
 *  With -v, prints all four corners and corresponding grid locations.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

void usage(char *progname);
void dofile(FILE *f, char *filename);
char *scan(FILE *f, char *filename);

int verbose = 0;	/* -v flag */
int exitcode = 0;	/* program exit code */



/*  main  */

int main(int argc, char *argv[])
{
    int c;
    FILE *f;
    char *filename, *msg;

    /* process options */

    while ((c = getopt(argc, argv, "v")) != -1) {
	switch (c) {
	    case 'v':
		verbose = 1;
		break;
	    default:
		usage(argv[0]);
	}
    }

    /* process file arguments */

    if (optind >= argc) {
	dofile(stdin, "[stdin]");
    } else {
	while (optind < argc) {
	    filename = argv[optind++];
	    f = fopen(filename, "rb");
	    if (f == NULL) {
		perror(filename);
		exitcode = 1;
	    } else {
		dofile(f, filename);
		fclose(f);
	    }
	}
    }

    return exitcode;
}


/*  dofile(file, filename) -- process one file  */

void dofile(FILE *f, char *filename)
{
    char *msg;

    msg = scan(f, filename);
    if (msg != NULL) {
	fprintf(stderr, "%s: %s\n", filename, msg);
	exitcode = 1;
    }
}



/*  scan(file, filename) -- scan DM file, printing vals or returning err msg  */

typedef struct { float x, y; } Coord;

#define BADFILE "not a DM file"

char *scan(FILE *f, char *filename)
{
    int c, i, w, h;
    char c1, c2, c3, c4;
    Coord a[8];

    c = getc(f);
    if (c == EOF)
	return "empty file";
    if (c != 'D' || getc(f) != 'M')
	return BADFILE;

    if (fscanf(f, " %d %d %c%c%c%c", &w, &h, &c1, &c2, &c3, &c4) != 6)
	return BADFILE;

    for (i = 0; i < 8; i++)
	if (fscanf(f, " %f %f", &a[i].x, &a[i].y) != 2)
	    return BADFILE;

    for (i = 0; i < 4; i++) {	/* scale arc-seconds to degrees */
	a[i].x /= 3600;
	a[i].y /= 3600;
    }

    if (verbose) {

	/* detailed output */
	printf("%-20s  ", filename);
	    for (i = 0; i < 4; i++)
	printf(" %7.2f %5.2f", a[i].x, a[i].y);
	printf("\n   %4d x %-4d        ", w, h);
	    for (i = 4; i < 8; i++)
	printf(" %6.1f %6.1f", a[i].x, a[i].y);
	printf("\n");

    } else {

	/* normal one-line output */
	printf("%4d %4d  %8.3f %6.3f  %8.3f %6.3f  %s\n",
	    w, h, a[0].x, a[0].y, a[2].x, a[2].y, filename);

    }

    return NULL;	/* no error */
}



/*  usage(progname) -- issue usage message and exit  */

void usage(char *progname)
{
    fprintf(stderr, "usage: %s [-v] [file...]\n", progname);
    exit(1);
}
