/*  emit.c -- draw a surface  */

#include "topo.h"
#include <GL/gl.h>

static void emitDFS(Hier *hp, TriIndex t, Coord2d p1, Coord2d p2, Coord2d p3);
static void emitColor(float x, float y, float z);
static void emitNormal(float,float,float,float,float,float,float,float,float);
static void emitTri(Hier *hp, TriIndex t, Coord2d p1, Coord2d p2, Coord2d p3);



static int surfType;		/* surface type: WIRE2D, WIRE3D, or FILL3D */



/*  emitSurface(hp, type) -- draw surface from triangle hierarchy  */

void emitSurface(Hier *hp, int type)
{
    surfType = type;
    emitDFS(hp, 1, se, nw, sw);
    emitDFS(hp, 2, nw, se, ne);
}



/*  emitDFS(hp, t, p1, p2, p3) -- emit triangle, possibly subdivided  */

static void emitDFS(Hier *hp, TriIndex t, Coord2d p1, Coord2d p2, Coord2d p3)
{
    Coord2d mid;

    if (ONSURFACE(hp, t)) {
	emitTri(hp, t, p1, p2, p3);
    } else {

	assert(!ISLEAF(hp, t));

	midpoint(p1, p2, mid);
	emitDFS(hp, LCHILD(t), p3, p1, mid);
	emitDFS(hp, RCHILD(t), p2, p3, mid);
    }
}



/*  emitTri(hp, t, p1, p2, p3) -- emit one triangle  */

static void emitTri(Hier *hp, TriIndex t, Coord2d p1, Coord2d p2, Coord2d p3)
{
    float x1, x2, x3, y1, y2, y3, z1, z2, z3;

    if (INROD(hp, t)) {

	if (surfType == WIRE2D) {

	    glBegin(GL_LINE_LOOP);
	    glVertex2i(p1[0], p1[1]);
	    glVertex2i(p2[0], p2[1]);
	    glVertex2i(p3[0], p3[1]);
	    glEnd();

	} else if (surfType == FILL2D) {

	    glBegin(GL_POLYGON);

	    switch(INFOV(hp, t)) {
	    case ALLINFOV:
		glColor3f(1.0, 1.0, 0.0);
		break;
	    case PARTINFOV:
		glColor3f(1.0, 0.0, 0.0);
		break;
	    default:
		glColor3f(0.5, 0.5, 0.5);
	    }

	    glVertex2i(p1[0], p1[1]);
	    glVertex2i(p2[0], p2[1]);
	    glVertex2i(p3[0], p3[1]);
	    glEnd();

	    glBegin(GL_LINE_LOOP);
	    glColor3f(0.0, 0.0, 0.0);
	    glVertex2i(p1[0], p1[1]);
	    glVertex2i(p2[0], p2[1]);
	    glVertex2i(p3[0], p3[1]);
	    glEnd();

	} else {	/* WIRE3D or FILL3D */

	    x1=p1[0];
	    y1=p1[1];
	    z1=vscale * dmRef(grid, p1[0], p1[1]);
	    x2=p2[0];
	    y2=p2[1];
	    z2=vscale * dmRef(grid, p2[0], p2[1]);
	    x3=p3[0];
	    y3=p3[1];
	    z3=vscale * dmRef(grid, p3[0], p3[1]);

	    if (surfType == WIRE3D) {
		glBegin(GL_LINE_LOOP);
		glVertex3f(x1, y1, z1);
		glVertex3f(x2, y2, z2);
		glVertex3f(x3, y3, z3);
		glEnd();
	    } else {
		glBegin(GL_POLYGON);
		emitNormal(x1, y1, z1, x2, y2, z2, x3, y3, z3);
		emitColor(x1, y1, z1); glVertex3f(x1, y1, z1);
		emitColor(x2, y2, z2); glVertex3f(x2, y2, z2);
		emitColor(x3, y3, z3); glVertex3f(x3, y3, z3);
		glEnd();
	    }
	}
    }
}



/*  emitNormal(x1, x2, x3, y1, y2, y3, z1, z2, z3) -- emit normal vector  */

static void emitNormal(float x1, float y1, float z1,
			float x2, float y2, float z2,
			float x3, float y3, float z3)
{
    /* We explicitly declare these (rather than putting them in a
     * Vector) so that they can be allocated into registers.
     */

    float px, py, pz, len;

    float v1x = x2 - x1,  v1y = y2 - y1,  v1z = z2 - z1;
    float v2x = x3 - x2,  v2y = y3 - y2,  v2z = z3 - z2;

    px = v1y * v2z - v1z * v2y;
    py = v1z * v2x - v1x * v2z;
    pz = v1x * v2y - v1y * v2x;

    /* Normalize */
    len = px * px + py * py + pz * pz;
    len = sqrt(len);
    px /= len;
    py /= len;
    pz /= len;

    glNormal3d(px, py, pz);

}



/*  emitColor(x, y, z) -- emit color depending on location and altitude  */

static void emitColor(float x, float y, float z)
{
    if (cgrid != NULL && reliefMethod == T_RGB){

	/* use color from PPM files */
	glColor3f(RED(cgrid, x, y) / 255.0,
		GREEN(cgrid, x, y) / 255.0,
		BLUE(cgrid, x, y) / 255.0);

    } else if (reliefMethod == T_CONTOURS){

	/* base the color on the elevation */
	Color *c;
	float sea;

	sea = vscale * SEA_LEVEL;
	if (z <= sea && z >= -sea) {
	    glColor3f(0.0, 0.0, 0.4);		/* use dark blue for sea level */
	} else {
	    c = elevMap + (int)(elevCount * (z - ZMIN) / (ZMAX - ZMIN + 1));
	    glColor3f(c->red, c->green, c->blue);
	}

    } else {

	glColor3f(0.6, 0.6, 0.55);		/* gray with slight tan tint */

    }
}
