/*  main.c -- TopoVista startup  */

#include <unistd.h>
#include <GL/gl.h>
#include <GL/glut.h>

#include "topo.h"
#include <sys/time.h>


/*  declare and initialize variables from globals.h  */
#define global
#define initial(v) = v
#include "globals.h"



/*  elevation color map, in ascending order */
/*  (also: sea level is dark blue, handled by special case in the code)  */

Color elevMap[] = {
    { 0.5, 0.5, 0.2 },	/* brown */
    { 0.3, 0.7, 0.3 },	/* green */
    { 0.3, 0.7, 0.3 },
    { 0.5, 0.9, 0.4 },	/* lighter green */
    { 0.5, 0.9, 0.4 },
    { 0.9, 0.9, 0.4 },	/* yellow */
    { 0.9, 0.9, 0.4 },
    { 0.9, 0.7, 0.4 },	/* orange */
    { 0.9, 0.7, 0.4 },
    { 0.8, 0.8, 0.7 },	/* light grayish tan */
    { 0.8, 0.8, 0.7 },
    { 1.0, 1.0, 1.0 },	/* white */
    { 1.0, 1.0, 1.0 },
};

int elevCount = sizeof(elevMap) / sizeof(elevMap[0]);	/* number of colors */



static void options(int argc, char *argv[]);
static void usage(char *progname);
static float mkscale(DMdata *grid);

static float vexagger = 1.0;		/* vertical exaggeration */



/*  main  */

void main(int argc, char **argv)
{
    int i;
    char *filename, *p;
    FILE *f;
    PPMdata *ppm;
    static long start,end;
    static struct timeval tval;

    glutInit(&argc, argv);		/* initialize GLUT system */

    options(argc, argv);		/* process command-line options */

    if (optind >= argc)			/* one filename argument is required */
	usage(argv[0]);

    /* read DM file */

    filename = argv[optind++];
    f = fopen(filename, "rb");
    if (f == NULL) {
	fprintf(stderr, "Can't open %s\n", filename);
	exit(1);
    }
    grid = dmRead(f);			/* read data */
    fclose(f);

    /* set dmname to basename of file for use in window titles */

    if ((p = strrchr(filename, '/')) == NULL)
	dmname = filename;
    else
	dmname = p + 1;

    i=strlen(dmname);
    if (i >= 3 && dmname[i-1] == 'm'
	       && dmname[i-2] == 'd'
	       && dmname[i-3] == '.')
        dmname[i-3] = '\0';

    /* read PPM files */
    if (optind < argc && (cgrid = initRGB(grid)) != NULL) {
	while (optind < argc) {
	    filename = argv[optind++];
	    f = fopen(filename, "r");
	    if (f == NULL) {
		fprintf(stderr, "Can't open %s\n", filename);
		exit(1);
	    }
	    ppm = ppmRead(f);
	    if (ppm) {
		mergeRGB(cgrid, ppm);
		free(ppm->red);
		free(ppm->green);
		free(ppm->blue);
		free(ppm);
	    }
	    fclose(f);

	}
    }

    vscale = mkscale(grid);		/* configure scaling */

    gettimeofday(&tval, NULL);
    start = tval.tv_sec;

    hierarchy = initHier();		/* initialize hierarchy */

    gettimeofday(&tval, NULL);
    fprintf(stderr,"Hierarchy built in %ld secs\n", tval.tv_sec - start);
    start = tval.tv_sec;


    /* initialize viewing parameters */
    eyeWx = 0.0;
    eyeWy = 0.0;
    refWx = 0.5 * XMAX;
    refWy = 0.5 * YMAX;

    /* open windows */
    openCtlWin();			/* control window */
    openOvhWin();			/* overhead (2-D) window */
    openPspWin();			/* perspective (3-D) window */

    /* process events */
    glutMainLoop();

    /*NOTREACHED*/
}



/*  options(argc, argv) -- interpret command-line arguments  */

static void options(int argc, char *argv[])
{
    int c;

    while ((c = getopt(argc, argv, "d:e:n:v:")) != -1) {
	switch (c) {
	    case 'd':
		stepThresh = atof(optarg);
		surfMethod = DIST;
		break;
	    case 'e':
		errThresh = atof(optarg);
		surfMethod = ERR;
		break;
	    case 'n':
		heapThresh = atoi(optarg);
		surfMethod = HEAP;
		break;
	    case 'v':
		vexagger = atof(optarg);
		break;
	    default:
		usage(argv[0]);
	}
    }
}



/*  usage(progname) -- issue usage message, and abort  */

static void usage(char *progname)
{
    fprintf(stderr,
	"usage: %s [-{denv} value]... DMfile [PPMfile ...]\n", progname);
    fprintf(stderr,
	"  -d <float>  distance threshold (for distance based surfaces)\n");
    fprintf(stderr,
	"  -e <float>  error threshold (for error based surfaces)\n");
    fprintf(stderr,
	"  -n <int>    max triangle count (for heap based surfaces)\n");
    fprintf(stderr,
	"  -v <float>  vertical scaling factor (default=1.0)\n");
    exit(1);
}



/*  mkscale(grid) -- configure scaling  */
/*  sets xmeters, ymeters, vscale  */

static float mkscale(DMdata *g)
{
    float a, cv, du, dv, xscale, yscale, vscale;

    /* get center latitude, in radians */
    cv = (g->nwUV.v + g->swUV.v + g->neUV.v + g->seUV.v) / 4;  /* arc-sec */
    a = cv / ARCSEC_PER_RAD;

    /* get dimensions of grid, in arc seconds */
    du = (g->neUV.u - g->nwUV.u + g->seUV.u - g->swUV.u) / 2;
    dv = (g->nwUV.v - g->swUV.v + g->neUV.v - g->seUV.v) / 2;

    /* get dimensions in meters */
    xmeters = du * METERS_PER_ARCSEC * cos(a);	/* scales by latitude */
    ymeters = dv * METERS_PER_ARCSEC;		/* always the same */

    /* get meters per grid cell */
    xscale = xmeters / g->width;
    yscale = ymeters / g->height;

    /* use the average of these to figure the vertical scale */
    vscale = vexagger / ((xscale + yscale) / 2);
    return vscale;
}
