/*  neighbor.c -- find neighbors of triangle  */

#include "topo.h"

/* The following chart describes the neighbor calculations in a
// recursive fashion: (Note: null concatenated with anything is null)
//
// N1(null) = N1(0) = N1(1) = null
// N1(w0) = N3(w)1
// N1(w1) = w0
//
// N2(null) = N2(0) = N2(1) = null
// N2(w0) = w1
// N2(w1) = N3(w)0
//
// N3(null) = null
// N3(0) = 1
// N3(1) = 0
// N3(w0) = N2(w)1
// N3(w1) = N1(w)0
*/

int neighbor[3][3] = {{0, 0, 0}, {0, 0, 0}, {0, 2, 1}};

#define SIEVE0 0xAAAAAAAA
#define SIEVE1 0x55555555



/*  cuzz(t, nbr) -- return the neighbor at the same level as t. */

static TriIndex cuzz(TriIndex t, int nbr)
{
    register unsigned int a;
    register unsigned int b;
    register unsigned int c;
    register unsigned int p;
    register unsigned int n;

    if (t <= 2) {
	return neighbor[nbr-1][t];
    }
    p = t + 1;
    if (nbr == 1) {
	p = (p << 1) | 1;
    } else if (nbr == 2) {
	p = (p << 1);
    }

    c = ((p << 1) ^ p) & SIEVE0;
    a = c | (c << 1) | 1;
    b = (a + 1) ^ a;
    if (b < (p>>1)) {
	n = p ^ b;
    } else if (((p & SIEVE0) << 1) > p) {
	if (((p & SIEVE1) << 2) > p) {
	    n = p ^ (b>>1);
	} else {
	    n = p ^ (b>>3);
	}
    } else {
	return 0;
    }

    if (nbr != 3) {
	return (n >> 1) - 1;
    } else {
	return n - 1;
    }
}



/*  cuz(t, nbr) -- return the neighbor at the same level as t. */

static TriIndex cuz(TriIndex t, int nbr)
{
    register int flipbit;
    register int nb;
    register int p;
    register int n;

    if (t) {
	p = t + 1;
	n = p;
	nb = nbr - 1;
	flipbit = 1;
	while (p > 3) {
	    n ^= flipbit;
	    nb ^= (p & 1) ^ (nb >> 1) ^ 2;
	    if (nb == 3) {
		return n - 1;
	    }
	    p >>= 1;
	    flipbit <<= 1;
	}
	if (nb == 2) {
	    return (n ^ flipbit) - 1;
	}
    }
    return 0;
}


static TriIndex cuz3(TriIndex t);

static TriIndex cuz1(TriIndex t)
/* Returns index of neighbor opposite vertex 1 of t at same level of tree */
{
    TriIndex n;

    if (t) {
	if (t & 1) {
	    n = cuz3(PARENT(t));
	    if (n) {
		return RCHILD(n);
	    } else {
		return 0;
	    }
	} else {
	    n = PARENT(t);
	    if (n) {
		return LCHILD(n);
	    } else {
		return 0;
	    }
	}
    } else {
	return 0;
    }
}

static TriIndex cuz2(TriIndex t)
/* Returns index of neighbor opposite vertex 2 of t at same level of tree */
{
    TriIndex n;

    if (t) {
	if (t & 1) {
	    n = PARENT(t);
	    if (n) {
		return RCHILD(n);
	    } else {
		return 0;
	    }
	} else {
	    n = cuz3(PARENT(t));
	    if (n) {
		return LCHILD(n);
	    } else {
		return 0;
	    }
	}
    } else {
	return 0;
    }
}

static TriIndex cuz3(TriIndex t)
/* Returns index of neighbor opposite vertex 3 of t at same level of tree */
{
    TriIndex n;

    if (t) {
	if (t == 1) {
	    return 2;
	}
	if (t == 2) {
	    return 1;
	}
	if (t & 1) {
	    n = cuz2(PARENT(t));
	    if (n) {
		return RCHILD(n);
	    } else {
		return 0;
	    }
	} else {
	    n = cuz1(PARENT(t));
	    if (n) {
		return LCHILD(n);
	    } else {
		return 0;
	    }
	}
    } else {
	return 0;
    }
}



TriIndex nbr1(Hier *hp, TriIndex t)
/* Returns index of neighbor opposite vertex 1 of t */
{
    TriIndex n;

    n = cuzz(t, 1);
    assert(n == cuz(t, 1));
    /*  n = cuz1(t);*/
    if (n && SMALLN1(hp, t)) {
	return LCHILD(n);
    } else {
	return n;
    }
}

TriIndex nbr2(Hier *hp, TriIndex t)
/* Returns index of neighbor opposite vertex 2 of t */
{
    TriIndex n;

    n = cuzz(t, 2);
    assert(n == cuz(t, 2));
    /*  n = cuz2(t); */
    if (n && SMALLN2(hp, t)) {
	return RCHILD(n);
    } else {
	return n;
    }
}

TriIndex nbr3(Hier *hp, TriIndex t)
/* Returns index of neighbor opposite vertex 3 of t */
{
    TriIndex n;

    n = cuzz(t, 3);
    assert(n == cuz(t, 3));
    /*  n = cuz3(t); */
    if (n && BIGN3(hp, t)) {
	return PARENT(n);
    } else {
	return n;
    }
}
