/*  overhead.c -- creation, drawing, and maintenance of overhead view window  */

#include <GL/gl.h>
#include <GL/glut.h>
#include "topo.h"
#include <sys/time.h>

void refreshOvh(void);
static void emitRelief(void);
static void emitViewfan(void);
static unsigned char *reliefImage(int w, int h);
static void mouseto(int btn, int x, int y);
static void mousefunc(int button, int state, int x, int y);
static void motionfunc(int x, int y);



static int whichBtn;			/* which mouse button is down? */



#define CMUL (255 * COLORATION)		/* color multiplier */
#define RMUL (255 * (1 - COLORATION))	/* relief multiplier */



/*  openOvhWin() -- initialize relief map window  */

void openOvhWin(void)
{
    char wname[200];
    int h = HEIGHT_2D;
    int w = HEIGHT_2D * xmeters / ymeters;

    sprintf(wname, "%s:O", dmname);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
    glutInitWindowSize(w, h);
    ovhWin = glutCreateWindow(wname);
    glutKeyboardFunc(keyfunc);
    glutSpecialFunc(specfunc);
    glutMouseFunc(mousefunc);
    glutMotionFunc(motionfunc);
    glutDisplayFunc(refreshOvh);
    if (cgrid)
	reliefMethod = T_RGB;
    else
	reliefMethod = T_CONTOURS;
    buildSurface();			/* ensure valid triangle count */
}



/*  refreshOvh() -- refresh overhead display  */

void refreshOvh(void)
{
    char buf[20];
    char wname[200];

    if (twoDMesh != FILL3D) {
	emitRelief();
	emitViewfan();
    }

    if (twoDMesh == NONE2D) {
        sprintf(wname, "%s:O", dmname);
	glutSetWindowTitle(wname);
    } else {
	buildSurface();
	sprintf(buf, "%d triangles", hierarchy->numSurfTri);
	glutSetWindowTitle(buf);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gluOrtho2D(0.0, XMAX - 1, 0.0, YMAX - 1);
	glColor3f(0.0, 0.0, 0.0);
	emitSurface(hierarchy, twoDMesh);
    }

    glutSwapBuffers();
}



/*  emitRelief() -- display relief map  */

static void emitRelief(void)
{
    int w, h;
    static int wprev, hprev, mprev;
    static unsigned char *a;

    /* get current window size */
    w = glutGet(GLUT_WINDOW_WIDTH);
    h = glutGet(GLUT_WINDOW_HEIGHT);

    /* reconstruct image only if size or mode has changed */
    if (w != wprev || h != hprev || reliefMethod != mprev)  {
	free(a);
	a = reliefImage(w, h);
	wprev = w;
	hprev = h;
	mprev = reliefMethod;
    }

    /* draw the image */
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glDrawPixels(w, h, GL_RGB, GL_UNSIGNED_BYTE, a);
}



/*  emitViewfan() -- draw lines indicating viewpoint and direction  */

static void emitViewfan(void)
{
    int w, h;
    float ex, ey, rx, ry, dx, dy, d, a, a1, a2;

    /* get current window size */
    w = glutGet(GLUT_WINDOW_WIDTH);
    h = glutGet(GLUT_WINDOW_HEIGHT);

    /* place eye and reference points in window coordinates */
    ex = eyeWx / XMAX * w;
    ey = eyeWy / YMAX * h;
    rx = refWx / XMAX * w;
    ry = refWy / YMAX * h;

    /* calculate distance and angles */
    dx = rx - ex;
    dy = ry - ey;
    d = sqrt(dx * dx + dy * dy);
    a = atan2(dy, dx);
    a1 = a + 0.35;
    a2 = a - 0.35;

    /* draw lines */
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    gluOrtho2D(0.0, w, 0.0, h);
    glColor3f(0.0, 0.0, 0.0);
    glBegin(GL_LINE_STRIP);
    glVertex2f(ex + d * cos(a1), ey + d * sin(a1));
    glVertex2f(ex, ey);
    glVertex2f(ex + d * cos(a2), ey + d * sin(a2));
    glEnd();
}



/*  reliefImage(w, h) -- construct relief image for w-by-h window */

static unsigned char *reliefImage(int w, int h)
{
    int *buf, *p;
    unsigned char *q;
    int i, j, k, x, y0, y1, d0, d1, z;
    int dz, dzmin, dzmax;
    float dx, dy, m, zm, lum, r, g, b;
    Color *c;
    UV s;

    /* allocate buffer: used first for ints, then overlaid with rgb bytes */
    buf = malloc(w * h * sizeof(int));
    if (!buf)  {
	fprintf(stderr, "Can't allocate space for relief image\n");
	exit(1);
    }

    /* compute step sizes and initial scaling factor */
    dx = grid->width / (float) w;
    dy = grid->height / (float) (h + 1);

    /* compute elevation difference for each output pixel */
    dzmin = dzmax = 0;
    p = buf;
    for (i = 0; i < h; i++) {
	y1 = grid->width * (int)(dy * (i + 1));
	if (dy < 1.0)
	    y0 = y1 - grid->width;
	else
	    y0 = y1 - grid->width * (int)(dy + 0.5);
	if (y0 < 0)
	    y0 = 0;
	for (j = 0; j < w; j++) {
	    x = dx * j;
	    d0 = grid->data[y0 + x];
	    d1 = grid->data[y1 + x];
	    dz = d0 - d1;
	    *p++ = dz;
	    /* find minima/maxima, omitting edges (which have min data value) */
	    if (d0 != EMIN && d1 != EMIN)  {
		if (dzmin > dz)
		    dzmin = dz;
		if (dzmax < dz)
		    dzmax = dz;
	    }
	}
    }

    /* calculate multiplier for scaling to [-0.5, +0.5] */
    if (dzmax > -dzmin)
	m = 0.5 / dzmax;
    else if (dzmin == 0)
	m = 1.0;
    else
	m = -0.5 / dzmin;

    /* stretch contrast beyond that */
    m *= CONTRAST;

    /* calculate elevation multiplier */
    zm = (float)elevCount / (EMAX - EMIN + 1);

    /* compute r, g, b triples, overlaying atop old int values */
    p = (int *) buf;
    q = (unsigned char *) buf;

    for (i = 0; i < h; i++) {
	y1 = grid->width * (int)(dy * (i + 1));  /* y offset in input grid */
	for (j = 0; j < w; j++) {

	    lum = 0.5 + m * *p++;		/* luminance */
	    if (lum > 1.0)
		lum = 1.0;			/* clamp to valid range */
	    if (lum < 0.0)
		lum = 0.0;

	    if (cgrid != NULL && reliefMethod == T_RGB) {
		s.u = j * dx;
		s.v = i * dy;
		r = RED(cgrid, s.u, s.v) / 255.0;
		g = GREEN(cgrid, s.u, s.v) / 255.0;
		b = BLUE(cgrid, s.u, s.v) / 255.0;

	    } else if (reliefMethod == T_CONTOURS) {
		x = dx * j;			/* x offset in input grid */
		z = grid->data[y1+x];		/* elevation */
		if (z <= SEA_LEVEL && z >= -SEA_LEVEL) {
		    r = 0.0;			/* use dark blue at sea level */
		    g = 0.0;
		    b = 0.4;
		} else {
		    k = zm * (z - EMIN);	/* color index */
		    c = &elevMap[k];		/* color entry */
		    r = c->red;
		    g = c->green;
		    b = c->blue;
		}

	    } else {
		r = g = b = lum;		/* use luminance only */
	    }

	    *q++ = RMUL * lum + CMUL * r;	/* r */
	    *q++ = RMUL * lum + CMUL * g;	/* g */
	    *q++ = RMUL * lum + CMUL * b;	/* b */
	}
    }

    /* discard unused portion of buffer, and return rgb bytes */
    return realloc(buf, 3 * w * h);
}



/*  mousefunc(btn, state, x, y) -- handle mouse click in overhead win */

static void mousefunc(int button, int state, int x, int y)
{
    whichBtn = button;			/* remember button for motion events */
    if (state != GLUT_UP)
	mouseto(button, x, y);		/* set position */
}



/*  motionfunc(x, y) -- handle mouse dragging  */

static void motionfunc(int x, int y)
{
    mouseto(whichBtn, x, y);
}



/*  mouseto(btn, x, y) -- handle mouse click or drag  */

static void mouseto(int btn, int x, int y)
{
    float Wx, Wy;

    Wx = XMAX * (float)x / glutGet(GLUT_WINDOW_WIDTH);
    Wy = YMAX * (1.0 - (float)y / glutGet(GLUT_WINDOW_HEIGHT));
					/* Y is reported as distance from top */

    if (btn == GLUT_RIGHT_BUTTON) {
	refWx = Wx;
	refWy = Wy;
	viewChanged();			/* note redisplay needed */

    } else if (btn == GLUT_LEFT_BUTTON) {
	eyeWx = Wx;
	eyeWy = Wy;
	viewChanged();			/* note redisplay needed */

    } else {
	printf("posn %8.3f %8.3f elev %8.3f\n", Wx, Wy, (double)zval(Wx, Wy));
    }

}

int doingLoop = 0;
int noRefresh = 0;

void doloop()
{
  int x,y,width,height;
  struct timeval tval;
  long start;

  glutSetWindow(ovhWin);
  width = glutGet(GLUT_WINDOW_WIDTH);
  height = glutGet(GLUT_WINDOW_HEIGHT);
  fprintf(stderr,"Window dimensions = %d x %d\n",width,height);
  x = 0;
  y = height-1;

  doingLoop = 1;

  gettimeofday(&tval, NULL);
  start = tval.tv_sec;

  for ( ;x < width; ++x)
    mouseto(GLUT_LEFT_BUTTON,x,y);

  for ( ;y >= 0; --y)
    mouseto(GLUT_LEFT_BUTTON,x,y);
    
  for ( ;x >= 0; --x)
    mouseto(GLUT_LEFT_BUTTON,x,y);
    
  for ( ;y < height; ++y)
    mouseto(GLUT_LEFT_BUTTON,x,y);

  gettimeofday(&tval, NULL);
  if (noRefresh) {
    fprintf(stderr,"Loop without refresh in %ld secs\n",tval.tv_sec-start);
  } else {
    fprintf(stderr,"Loop with refresh in %ld secs\n",tval.tv_sec-start);
  }

  doingLoop = 0;
}
    
