/*  ppmops.c -- operations on PPM files  */

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include "ppmops.h"

static Transform *dmxy2ll;
static UV dmswUV, dmnwUV, dmneUV, dmseUV;


void discardToCR(FILE *in)
{
    int c;

    while ((c = getc(in)) != EOF && c != '\n')
	;
}


void ppmGetToken(FILE *in, char *tok, int n)
{
    int c, i;

    while ((c = getc(in)) != EOF && (isspace(c) || c == '#')) {
	if (c == '#') {
	    discardToCR(in);
	}
    }
    tok[0] = c;
    i = 1;
    while (i < n && (c = getc(in)) != EOF && ! isspace(c) && c != '#') {
	tok[i++] = c;
    }
    ungetc(c, in);
    tok[i] = '\0';
}


PPMdata *ppmRead(FILE *in)
{
    PPMdata *ppm;
    long i;
    int max;
    char buf[100];
    int row, col;
    unsigned char *pr, *pg, *pb;


    ppm = malloc(sizeof(PPMdata));
    if (ppm) {
	ppm->ll2xy = NULL;
	i = ftell(in);
	fscanf(in, "P6\n#TopoVista1\n");	/* Required */
	i = ftell(in) - i;
	if (i == 15) {
	    fscanf(in, "#lon\t\tlat\t\tx\t\ty\n");	/* Optional */
	    if (fscanf(in,			/* Required */
"# %lf %lf %lf %lf\n# %lf %lf %lf %lf\n# %lf %lf %lf %lf\n# %lf %lf %lf %lf\n",
		&(ppm->swUV.u),&(ppm->swUV.v),&(ppm->swXY.u),&(ppm->swXY.v),
		&(ppm->nwUV.u),&(ppm->nwUV.v),&(ppm->nwXY.u),&(ppm->nwXY.v),
		&(ppm->neUV.u),&(ppm->neUV.v),&(ppm->neXY.u),&(ppm->neXY.v),
		&(ppm->seUV.u),&(ppm->seUV.v),&(ppm->seXY.u),&(ppm->seXY.v))
		== 16) {
	            ppm->swUV.u *= 3600.;	/* Convert to arcseconds */
	            ppm->swUV.v *= 3600.;
	            ppm->nwUV.u *= 3600.;
	            ppm->nwUV.v *= 3600.;
	            ppm->neUV.u *= 3600.;
	            ppm->neUV.v *= 3600.;
	            ppm->seUV.u *= 3600.;
	            ppm->seUV.v *= 3600.;

		    ppm->ll2xy = createTransform(
			ppm->swUV, ppm->nwUV, ppm->neUV, ppm->seUV,
			ppm->swXY, ppm->nwXY, ppm->neXY, ppm->seXY);
	    }
	}

	if (i >=4 && i < 15) {
	    discardToCR(in);
	}
	if (i >= 2) {
	    ppmGetToken(in, buf, 100);
	    ppm->width = atoi(buf);
	    ppmGetToken(in, buf, 100);
	    ppm->height = atoi(buf);
	    ppmGetToken(in, buf, 100);
	    max = atoi(buf);

	    fprintf(stderr, " ppm size %d x %d\n", ppm->width, ppm->height);

	    getc(in);			/* discard \n */

	    ppm->red = malloc(ppm->width * ppm->height);
	    ppm->green = malloc(ppm->width * ppm->height);
	    ppm->blue = malloc(ppm->width * ppm->height);
	    if (ppm->red && ppm->green && ppm->blue) {

		/*
		 * PPM files are stored starting at NW corner while DM files
		 * (and all structures in this program) start at SW corner.
		 * So we have to flip the ppm file over.
		 */
		for(row = ppm->height - 1; row >= 0; row--) {
		    pr = ppm->red + row * ppm->width;
		    pg = ppm->green + row * ppm->width;
		    pb = ppm->blue + row * ppm->width;
		    for(col = 0; col < ppm->width; col++) {
			*pr++ = 255. * (float)getc(in) / (float)max;
			*pg++ = 255. * (float)getc(in) / (float)max;
			*pb++ = 255. * (float)getc(in) / (float)max;
		    }
		}
		return ppm;
	    } else {
		fprintf(stderr,
		    "ppmRead: Unable to allocate memory for rgb arrays.\n");
	    }
	} else {
	    fprintf(stderr, "Cannot read ppm file (wrong format)\n");
	}
    } else {
	fprintf(stderr, "ppmRead: Unable to allocate memory.\n");
    }
    free(ppm->red);
    free(ppm->green);
    free(ppm->blue);
    free(ppm);
    return NULL;
}

void ppmWriteHeader(FILE *out, int width, int height,
		    UV swUV, UV nwUV, UV neUV, UV seUV,
		    UV swXY, UV nwXY, UV neXY, UV seXY)
{
    fprintf(out, "P6\n");
    fprintf(out, "#TopoVista1\n");
    fprintf(out, "#lon\t\tlat\t\tx\t\ty\n");
    fprintf(out, "#%lf\t%lf\t%lf\t%lf\n#%lf\t%lf\t%lf\t%lf\n#%lf\t%lf\t%lf\t%lf\n#%lf\t%lf\t%lf\t%lf\n",
		    swUV.u/3600., swUV.v/3600.,swXY.u, swXY.v,
		    nwUV.u/3600., nwUV.v/3600.,nwXY.u, nwXY.v,
		    neUV.u/3600., neUV.v/3600.,neXY.u, neXY.v,
		    seUV.u/3600., seUV.v/3600.,seXY.u, seXY.v);
		    
    fprintf(out, "%d %d\n", width, height);
    fprintf(out, "255\n");
}



RGBdata *initRGB(DMdata *dm)
{
    RGBdata *rgb;
    int x, y;
    unsigned char *red, *green, *blue;

    rgb = malloc(sizeof(RGBdata));
    if (rgb) {
	rgb->red = malloc(dm->width * dm->height);
	rgb->green = malloc(dm->width * dm->height);
	rgb->blue = malloc(dm->width * dm->height);
    }
    if (!rgb || !rgb->red || !rgb->green || !rgb->blue) {
	fprintf(stderr, "Not enough memory to create RGB array\n");
	free(rgb);
	return NULL;
    }
    rgb->width = dm->width;
    rgb->height = dm->height;
    rgb->number_of_layers = 0;

    red = rgb->red;
    green = rgb->green;
    blue = rgb->blue;
    for (y = 0; y < rgb->height; ++y) {
	for(x = 0; x < rgb->width; ++x) {
	    *red++ = 0.4;
	    *green++ = 0.4;
	    *blue++ = 0.4;
	}
    }

    dmxy2ll = createTransform(dm->swXY, dm->nwXY, dm->neXY, dm->seXY,
			    dm->swUV, dm->nwUV, dm->neUV, dm->seUV);

    dmswUV = dm->swUV;
    dmnwUV = dm->nwUV;
    dmneUV = dm->neUV;
    dmseUV = dm->seUV;

    return rgb;
}


void mergeRGB(RGBdata *rgb, PPMdata *ppm)
{
    int x, y, n;
    unsigned char *red, *green, *blue;
    Transform *ppmll2xy;
    UV swXY, nwXY, neXY, seXY;
    UV s, t;
    if (ppm->ll2xy) {
	ppmll2xy = ppm->ll2xy;
    } else {		/* use default trans */
	fprintf(stderr, "Assuming ppm file covers dm file\n");
	swXY.u = 0;
	swXY.v = 0;
	nwXY.u = 0;
	nwXY.v = ppm->height;
	neXY.u = ppm->width;
	neXY.v = ppm->height;
	seXY.u = ppm->width;
	seXY.v = 0;
	ppmll2xy = createTransform(dmswUV, dmnwUV, dmneUV, dmseUV,
				    swXY, nwXY, neXY, seXY);
    }

    red = rgb->red;
    green = rgb->green;
    blue = rgb->blue;

    n = rgb->number_of_layers;
    for(y = 0; y < rgb->height; ++y) {
	for(x = 0; x < rgb->width; ++x) {
	    s.u = x;
	    s.v = y;

	    transform(dmxy2ll, s, &t);
	    transform(ppmll2xy, t, &s);

	    /* doesn't do the right thing if, for example, layer 1 assigns
	     * no color to x,y and layer 2 assigns black to x,y.  The result,
	     * currently, is an even mix of black and (0.4, 0.4, 0.4) rather
	     * than black.
	     */

	    if (s.u < ppm->width && s.u >= 0 && s.v < ppm->height && s.v >= 0) {
		*red = (*red * n + PPMRED(ppm, s.u, s.v)) / (n + 1);
		*green = (*green * n + PPMGREEN(ppm, s.u, s.v)) / (n + 1);
		*blue = (*blue * n + PPMBLUE(ppm, s.u, s.v)) / (n + 1);
	    }
	    ++red;
	    ++green;
	    ++blue;
	}
    }
    rgb->number_of_layers += 1;
}
