/*  surface.c -- compute surface based on error criteria  */

#include "topo.h"


#define LEFT_TURN(a, b, c) \
    (((b)[1]-(a)[1])*((c)[0]-(a)[0])-((b)[0]-(a)[0])*((c)[1]-(a)[1]) < 0)


static Coord2d eyeP;
static int scaledErrThresh;

static int depth(TriIndex t)
{
    int d = 0;
    int p = t + 1;

    while (p > 1) {
	d += 1;
	p >>= 1;
    }
    return d;
}


/* badness(hp,t,Coord2d p1, Coord2d p2, Coord2d p3)
 *	 -- returns integer measure of triangle's badness from -1 to MAXBAD
 */

static int badness(Hier *hp, TriIndex t, Coord2d p1, Coord2d p2, Coord2d p3)
{
    int h = hp->node[t].err;
    int dx = p3[0] - eyeP[0];
    int dy = p3[1] - eyeP[1];
    int d = dx * dx + dy * dy;

    if (!INFOV(hp, t))			/* if not in field of view */
	return -1;

    if (d == 0)				/* if right below eye */
	return MAXBAD;

    /* in DISTANCE mode, check distance from eye */
    if (surfMethod == DIST
    && (double)(hp->depth - depth(t)) - DILATION * log((double)d) > stepThresh)
	return MAXBAD;			/* if close enough to be needed */

    d = (int)sqrt((double)d);
    dx = h * errScale * vscale * MAXBAD / d;

    return MIN(dx, MAXBAD);

}

static void rodSplit(Hier *hp, TriIndex t)
{
    if (!ISLEAF(hp, t) && BORDER(hp, t)) {
	if (ONSURFACE(hp, t)) {
	    split(hp, t);
	}
	rodSplit(hp, LCHILD(t));
	rodSplit(hp, RCHILD(t));
    }
}


static void eyeSplit(Hier *hp, TriIndex t, Coord2d p1, Coord2d p2, Coord2d p3)
{
    Coord2d mid;

    if (!ISLEAF(hp, t)) {
	if (ONSURFACE(hp, t)) {
	    split(hp, t);
	}
	midpoint(p1, p2, mid);
	if (LEFT_TURN(mid, p3, eyeP)) {
	    eyeSplit(hp, LCHILD(t), p3, p1, mid);
	    rodSplit(hp, RCHILD(t));
	} else {
	    eyeSplit(hp, RCHILD(t), p2, p3, mid);
	    rodSplit(hp, LCHILD(t));
	}
    }
}

/* baseSurface --
 * calculates surface with finest level of detail around eye position
 * and no triangles spanning border of rectangle of definition.
 * This is more efficient than checking in badness() if triangle
 * contains the eye point.
 */
static void baseSurface(Hier *hp)
{
    eyeP[0] = (int)eyeWx;
    eyeP[1] = (int)eyeWy;

    PUTONSURFACE(hp, 1);    /* SURFACE starts as triangles 1 and 2 */
    PUTONSURFACE(hp, 2);
    hp->node[1].info |= triInFOV(1);
    hp->node[2].info |= triInFOV(2);

    hp->numSurfTri = 2;

    if (LEFT_TURN(se, nw, eyeP)) {
	eyeSplit(hp, 1, se, nw, sw);
	rodSplit(hp, 2);
    } else {
	eyeSplit(hp, 2, nw, se, ne);
	rodSplit(hp, 1);
    }
}


/*-------------------ERROR SURFACE-----------------------------*/

static int tooCoarse(Hier *hp, TriIndex t, Coord2d p1, Coord2d p2, Coord2d p3)
{

    if (badness(hp, t, p1, p2, p3) > errThresh * MAXBAD)
	return 1;
    else
	return 0;
}


static void errSplit(Hier *hp, TriIndex t, Coord2d p1, Coord2d p2, Coord2d p3)
{
    Coord2d mid;

    if (!ISLEAF(hp, t) && tooCoarse(hp, t, p1, p2, p3)) {
	if (ONSURFACE(hp, t)) {
	    split(hp, t);
	}
	midpoint(p1, p2, mid);
	errSplit(hp, LCHILD(t), p3, p1, mid);
	errSplit(hp, RCHILD(t), p2, p3, mid);
    }
}


void errSurface(Hier *hp)
{
    baseSurface(hp);
    errSplit(hp, 1, se, nw, sw);
    errSplit(hp, 2, nw, se, ne);
}

/*--------------------HEAP SURFACE------------------------------*/

/* heapBadness(hp,t) -- return integer badness of triangle t */

int heapBadness(Hier *hp, TriIndex t)
{
    Coord2d p1, p2, p3;

    getVerts(t, p1, p2, p3);
    return badness(hp, t, p1, p2, p3);
}



/* worstTri(hp) -- return highest heap item that is ONSURFACE */

static TriIndex worstTri(Hier *hp)
{
    TriIndex t;

    t = heapExtractMax();
    while (t != 0 && (!ONSURFACE(hp, t) || !INFOV(hp, t))) {
	t = heapExtractMax();
    }
    return t;
}


/*  heapSurface(hp) -- compute surface using heap algorithm  */

void heapSurface(Hier *hp)
{
    TriIndex t;

    heapEmpty();
    heapInsert(1, heapBadness(hp, 1));
    heapInsert(2, heapBadness(hp, 2));
    baseSurface(hp);

    while (hp->numSurfTri < heapThresh || heapTopVal() == MAXBAD) {
	t = worstTri(hp);
	if (t == 0)
	    break;

	if (!ISLEAF(hp, t))
	    split(hp, t);
    }
}
