/*  transform.c -- coordinate system transformations  */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "basic.h"



/*  transform(f, s, t) -- transform t to s using transformation f  */

void transform(Transform *f, UV s, UV *t)
{
    t->u = f->ua + s.u * f->ub + s.v * f->uc + s.u * s.v * f->ud;
    t->v = f->va + s.u * f->vb + s.v * f->vc + s.u * s.v * f->vd;
}



/*  createTransform(swS,nwS,neS,seS,swT,nwT,neT,seT) -- create transformation */

Transform *createTransform(UV swS, UV nwS, UV neS, UV seS,
			    UV swT, UV nwT, UV neT, UV seT)
{
    Transform *f;
    double x1, y1, x2, y2, x3, y3, x4, y4;
    double s1, s2, s3, s4;
    double a11, a12, a13, a14, a21, a22, a23, a24;
    double a31, a32, a33, a34, a41, a42, a43, a44;
    double det;

    f = malloc(sizeof(Transform));
    if (f == NULL) {
	fprintf(stderr, "Out of space allocating transform\n");
	exit(1);
    }

    x1 = swS.u;
    y1 = swS.v;
    x2 = nwS.u;
    y2 = nwS.v;
    x3 = neS.u;
    y3 = neS.v;
    x4 = seS.u;
    y4 = seS.v;

    det = (x1 * x2 + x3 * x4) * (y1 - y2) * (y4 - y3)
	+ (x1 * x3 + x2 * x4) * (y1 - y3) * (y2 - y4)
	+ (x1 * x4 + x2 * x3) * (y1 - y4) * (y3 - y2);

    if (det == 0) {
	fprintf(stderr, "Transform singular\n");
	exit(1);
    }

#ifdef DEBUG
    fprintf(stderr, "Det %lf\n", det);
#endif

    a31 = x3 * x4 * (y4 - y3) + x2 * x4 * (y2 - y4) + x2 * x3 * (y3 - y2);
    a34 = x2 * x3 * (y2 - y3) + x1 * x3 * (y3 - y1) + x1 * x2 * (y1 - y2);
    a32 = x3 * x4 * (y3 - y4) + x1 * x4 * (y4 - y1) + x1 * x3 * (y1 - y3);
    a11 = x2 * y3 * x4 * y4 - x2 * x3 * y3 * y4 - x3 * y2 * x4 * y4
	+ x3 * x2 * y2 * y4 + x4 * y2 * x3 * y3 - x4 * x2 * y2 * y3;
    a44 = x2 * y3 - y2 * x3 - x1 * y3 + y1 * x3 + x1 * y2 - y1 * x2;
    a14 = - x1 * y2 * x3 * y3 + x1 * x2 * y2 * y3 + x2 * y1 * x3 * y3
	- x2 * x1 * y1 * y3 - x3 * y1 * x2 * y2 + x3 * x1 * y1 * y2;
    a33 = x2 * x4 * (y4 - y2) + x1 * x4 * (y1 - y4) + x1 * x2 * (y2 - y1);
    a12 =  - x1 * y3 * x4 * y4 + x1 * x3 * y3 * y4 + x3 * y1 * x4 * y4
	- x3 * x1 * y1 * y4 - x4 * y1 * x3 * y3 + x4 * x1 * y1 * y3;
    a21 = - y3 * x4 * y4 + x3 * y3 * y4 + y2 * x4 * y4
	- x2 * y2 * y4 - y2 * x3 * y3 + x2 * y2 * y3;
    a41 = - x3 * y4 + y3 * x4 + x2 * y4 - y2 * x4 - x2 * y3 + y2 * x3;
    a24 = y2 * x3 * y3 - x2 * y2 * y3 - y1 * x3 * y3
	+ x1 * y1 * y3 + y1 * x2 * y2 - x1 * y1 * y2;
    a42 = x3 * y4 - y3 * x4 - x1 * y4 + y1 * x4 + x1 * y3 - y1 * x3;
    a22 = y3 * x4 * y4 - x3 * y3 * y4 - y1 * x4 * y4
	+ x1 * y1 * y4 + y1 * x3 * y3 - x1 * y1 * y3;
    a43 = - x2 * y4 + y2 * x4 + x1 * y4 - y1 * x4 - x1 * y2 + y1 * x2;
    a23 = - y2 * x4 * y4 + x2 * y2 * y4 + y1 * x4 * y4
	- x1 * y1 * y4 - y1 * x2 * y2 + x1 * y1 * y2;
    a13 = x1 * y2 * x4 * y4 - x1 * x2 * y2 * y4 - x2 * y1 * x4 * y4
	+ x2 * x1 * y1 * y4 + x4 * y1 * x2 * y2 - x4 * x1 * y1 * y2;

#ifdef DEBUG
    fprintf(stderr, "%lf %lf %lf %lf\n", a11/det, a12/det, a13/det, a14/det);
    fprintf(stderr, "%lf %lf %lf %lf\n", a21/det, a22/det, a23/det, a24/det);
    fprintf(stderr, "%lf %lf %lf %lf\n", a31/det, a32/det, a33/det, a34/det);
    fprintf(stderr, "%lf %lf %lf %lf\n", a41/det, a42/det, a43/det, a44/det);
#endif

    s1 = swT.u;
    s2 = nwT.u;
    s3 = neT.u;
    s4 = seT.u;

    f->ua = (a11 * s1 + a12 * s2 + a13 * s3 + a14 * s4) / det;
    f->ub = (a21 * s1 + a22 * s2 + a23 * s3 + a24 * s4) / det;
    f->uc = (a31 * s1 + a32 * s2 + a33 * s3 + a34 * s4) / det;
    f->ud = (a41 * s1 + a42 * s2 + a43 * s3 + a44 * s4) / det;

#ifdef DEBUG
    fprintf(stderr, "U: %lf %lf %lf %lf\n", f->ua, f->ub, f->uc, f->ud);
#endif

    s1 = swT.v;
    s2 = nwT.v;
    s3 = neT.v;
    s4 = seT.v;

    f->va = (a11 * s1 + a12 * s2 + a13 * s3 + a14 * s4) / det;
    f->vb = (a21 * s1 + a22 * s2 + a23 * s3 + a24 * s4) / det;
    f->vc = (a31 * s1 + a32 * s2 + a33 * s3 + a34 * s4) / det;
    f->vd = (a41 * s1 + a42 * s2 + a43 * s3 + a44 * s4) / det;

#ifdef DEBUG
    fprintf(stderr, "V: %lf %lf %lf %lf\n", f->va, f->vb, f->vc, f->vd);
#endif

    return f;

}



/*  utm2ll(zone, utm, ll) -- transform UTM coordinates to latitude/longitude  */

/* This assumes the Clark 1866 ellipsoid.  See
 * Map Projections -- A Working Manual  by John P. Snyder
 * U.S. Geological Survey Professional Paper 1395
 */

void utm2ll(int zone, UV utm, UV *ll)
{
    double k0 = 0.9996;
    double a = 6378206.4;
    double es = 0.00676866;
    double lam0, x, y, M, mu, e1, phi1, eps, phi, lam, C1, T1, N1, R1, D;

    lam0 = ((double)zone - 0.5) * PI / 30.0 - PI;

    x = utm.u - 500000;
    y = utm.v;

    M = y / k0;
    mu = M / (a * (1 - es / 4. - 3. * es * es / 64. - 5. * es*es*es / 256.));
    e1 = (1 - sqrt(1 - es)) / (1 + sqrt(1 - es));


    phi1 = mu + (3. * e1 / 2. - 27. * e1 * e1 * e1 / 32.) * sin(2. * mu)
	    + (21. * e1 * e1 / 16. - 55. * e1*e1*e1*e1 / 32.) * sin(4. * mu)
	    + (151. * e1 * e1 * e1 / 96.) * sin(6. * mu)
	    + (1097. * e1 * e1 * e1 * e1 / 512.) * sin(8. * mu);

    eps = es / (1 - es);
    C1 = eps * cos(phi1) * cos(phi1);
    T1 = tan(phi1) * tan(phi1);
    N1 = a / sqrt(1 - es * sin(phi1) * sin(phi1));
    R1 = a*(1-es) / (sqrt(1-es*sin(phi1)*sin(phi1))*(1-es*sin(phi1)*sin(phi1)));
    D = x / (N1 * k0);

    phi = phi1 - N1 * tan(phi1 / R1) *
	(D * D / 2. - (5. + 3. * T1 + 10.*C1 - 4.*C1*C1 - 9.*eps)* D*D*D*D / 24.
	+ (61. + 90. * T1 + 298. * C1 + 45. * T1*T1 - 252. * eps - 3. * C1*C1)
	    * D * D * D * D * D * D / 720.);
    lam = lam0 + (D - (1 + 2 * T1 + C1) * D*D*D / 6. + (5. - 2. * C1 + 28. * T1
	- 3. * C1*C1 + 8. * eps + 24. * T1*T1) * D*D*D*D*D / 120.) / cos(phi1);


    ll->u = lam * ARCSEC_PER_RAD;
    ll->v = phi * ARCSEC_PER_RAD;
}
