/*  view.c -- display perspective view  */

#include "topo.h"
#include <GL/gl.h>
#include <GL/glut.h>

static void refreshPsp(void);


static int newSurface = 1;	/* does surface need to be recomputed? */

static float cosHalfAngle2 = 0.85355339;
static float angle = 45.0;
static float near = 0.1;
static float far;



/*  viewChanged() -- take note that view has changed, so surface will, too  */

void viewChanged(void)
{
  extern int doingLoop;
  extern int noRefresh;
  extern void refreshOvh();

  if (doingLoop) {
    newSurface = 1;
    if (noRefresh) {
      buildSurface();
    } else {
      glutSetWindow(ovhWin);
      refreshOvh();
      glutSetWindow(pspWin);
      refreshPsp();
      glutSetWindow(ovhWin);
    }
  } else {
    newSurface = 1;
    glutSetWindow(ovhWin);
    glutPostRedisplay();
    glutSetWindow(pspWin);
    glutPostRedisplay();
  }
}



/*  buildSurface() -- rebuild the surface if newSurface is set  */

void buildSurface(void)
{
    int x, y, z;

    if (!newSurface)		/* if nothing's changed, don't rebuild */
	return;

    /* compute eye height depending on location and eyeUp flag */

    if (eyeUp) {

	/* eye is up: constant height above highest peak */
	eyeWz = ZMAX + 0.5 * (ZMAX - ZMIN);

    } else {

	/* eye down: just above max height of grid cell corners */
	x = eyeWx;
	y = eyeWy;
	z = zval(x, y);
	z = MAX(z, zval(x + 1, y));
	z = MAX(z, zval(x + 1, y + 1));
	z = MAX(z, zval(x, y + 1));
	eyeWz = vscale * (z + EYE_HEIGHT);
    }

    /* compute reference height and vector */

    if (refWx < 0 || refWx >= XMAX || refWy < 0 || refWy >= YMAX)
	refWz = ZMIN;
    else
	refWz = vscale * (dmRef(grid, (int)refWx, (int)refWy) + REF_HEIGHT);

    Rx = refWx - eyeWx;
    Ry = refWy - eyeWy;
    Rz = refWz - eyeWz;
    inAngleThresh = (Rx * Rx + Ry * Ry) * cosHalfAngle2;

    /* build the surface */

    switch(surfMethod) {
	case HEAP:
	default:
	    heapSurface(hierarchy);
	    break;
	case DIST:
	case ERR:
	    errSurface(hierarchy);
	    break;
	}

    newSurface = 0;
}



/*  zval(x, y) -- return z value from grid, or EMIN if out of bounds  */

int zval(int x, int y)
{
    if (x < 0 || y < 0 || x >= grid->width || y >= grid->height)
	return EMIN;
    else
	return dmRef(grid, x, y);
}



/*  openPspWin() -- open perspective view window  */

void openPspWin(void)
{

    GLfloat light_specular[] = { 0.2, 0.2, 0.2, 1.0 };
    GLfloat light_diffuse[] = { 1.0, 1.0, 1.0, 1.0 };
    GLfloat light_ambient[] = { 0.0, 0.0, 0.0, 1.0 };
    GLfloat light_position[] = { 0.0, 1.0, 2.0, 0.0 };
    char wname[200];

    far = 2 * (DIMENSION + DIMENSION);

    glutInitWindowSize(WIDTH_3D, HEIGHT_3D);
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
    sprintf(wname, "%s:P", dmname);
    pspWin = glutCreateWindow(wname);
    glutSetWindowTitle(wname);

    glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
    glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
    glLightfv(GL_LIGHT0, GL_POSITION, light_position);
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);

    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);
    glEnable(GL_COLOR_MATERIAL);
    glDisable(GL_POLYGON_SMOOTH);

    glDepthFunc(GL_LEQUAL);
    glEnable(GL_DEPTH_TEST);
    glutDisplayFunc(refreshPsp);
    glutKeyboardFunc(keyfunc);
    glutSpecialFunc(specfunc);
}



/*  refreshPsp -- redraw perspective (3-D) view  */

static void refreshPsp(void)
{
    int w, h;
    float aspect;
    GLfloat light_position[4] = {0.0, 1.0, 2.0, 0.0};  /* high in northern sky*/

    buildSurface();

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    w = glutGet(GLUT_WINDOW_WIDTH);
    h = glutGet(GLUT_WINDOW_HEIGHT);
    aspect = (float)w / (float)h;
    gluPerspective((GLdouble)angle / aspect, (GLdouble)aspect,
		    (GLdouble)near, (GLdouble)far);
    gluLookAt((GLdouble)eyeWx, (GLdouble)eyeWy, (GLdouble)eyeWz,
		    (GLdouble)refWx, (GLdouble)refWy, (GLdouble)refWz,
		    (GLdouble)0.0, (GLdouble)0.0, (GLdouble)1.0);

    /* high northern light source matches the overhead relief view */
    glDisable(GL_LIGHTING);
    glDisable(GL_LIGHT0);
    glLightfv(GL_LIGHT0, GL_POSITION, light_position);
    glEnable(GL_LIGHT0);
    glEnable(GL_LIGHTING);

    glClearColor(0.0, 0.0, 1.0, 0.0);
    glClearDepth((GLclampd)far);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glColor3f(1.0, 1.0, 1.0);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    emitSurface(hierarchy, FILL3D);

    if (threeDMesh) {
	glColor3f(0.0, 0.0, 0.0);
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	emitSurface(hierarchy, WIRE3D);
    }

    glutSwapBuffers();

}
